#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var yargs = require('yargs');
var bodyParser = require('body-parser');
var express = require('express');
var chalk = _interopDefault(require('chalk'));
var webpack_ = require('webpack');
var fs = require('fs');
var fastGlob = require('fast-glob');
var path = require('path');
var merge = require('webpack-merge');
var requireFromString_ = require('require-from-string');

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var prefix = chalk.grey("λ-dev");
function getBuildCallback(entries, resolve, reject, callback) {
    return function (err, stats) {
        if (err) {
            console.error(chalk.grey("λ-dev") + " " + chalk.red("Webpack error:"), err);
            reject({ entries: entries, error: err });
        }
        var _a = stats.toJson({
            assets: true,
            errors: true,
            hash: true,
            warnings: true
        }), errors = _a.errors, hash = _a.hash, modules = _a.modules, warnings = _a.warnings;
        if (errors.length > 0) {
            for (var _i = 0, errors_1 = errors; _i < errors_1.length; _i++) {
                var error = errors_1[_i];
                console.error(chalk.grey("λ-dev") + " " + chalk.red("Compilation error:"), error);
            }
            reject({ entries: entries, errors: errors, hash: hash });
        }
        if (warnings.length > 0) {
            for (var _b = 0, warnings_1 = warnings; _b < warnings_1.length; _b++) {
                var warning = warnings_1[_b];
                console.warn(chalk.grey("λ-dev") + " " + chalk.red("Compilation warning:"), warning);
            }
        }
        console.log(prefix + " Built " + chalk.green(hash));
        if (callback) {
            callback({ entries: entries, hash: hash, modules: modules });
        }
        resolve({ entries: entries, hash: hash, modules: modules, warnings: warnings });
    };
}

var CWD = process.cwd();
var getGlobAndCwd = function (entry) {
    var resolved = path.resolve(entry);
    var dirStat = fs.statSync(resolved);
    if (dirStat.isDirectory()) {
        return { glob: "**/*", cwd: resolved };
    }
    else {
        var dir = path.parse(resolved).dir;
        return { glob: resolved, cwd: dir };
    }
};
function getEntries(entry, include, exclude) {
    if (entry === void 0) { entry = CWD; }
    if (exclude === void 0) { exclude = []; }
    try {
        var _a = getGlobAndCwd(entry), glob = _a.glob, cwd_1 = _a.cwd;
        var files = fastGlob.sync(include || glob, {
            absolute: true,
            cwd: entry,
            ignore: exclude,
            matchBase: true,
            onlyFiles: true,
            stats: false // return file path string
        });
        return files.map(function (file) {
            var parsed = path.parse(file);
            return {
                file: file,
                target: parsed.dir.replace(cwd_1, "") + "/" + parsed.base,
                requestPath: parsed.dir.replace(cwd_1, "") + "/" + parsed.name.replace(/^index$/, "")
            };
        });
    }
    catch (err) {
        console.error(chalk.grey("λ-dev") + " " + chalk.red("Glob error:"), err);
    }
}

var CWD$1 = process.cwd();
function getWebpackConfig(_a) {
    var customConfig = _a.customConfig, _b = _a.dev, dev = _b === void 0 ? false : _b, entries = _a.entries, nodeVersion = _a.nodeVersion, _c = _a.targetDir, targetDir = _c === void 0 ? "." : _c;
    var path$$1 = path.resolve(CWD$1, targetDir);
    var entry = entries.reduce(function (accumulator, entry) {
        var key = entry.requestPath.replace(/^\//, "");
        accumulator[key] = entry.file;
        return accumulator;
    }, {});
    var defaultConfig = {
        mode: dev ? "development" : "production",
        watch: dev,
        entry: entry,
        output: { filename: "[name].js", libraryTarget: "commonjs", path: path$$1 },
        target: "node",
        module: {
            rules: [
                {
                    test: /\.js$/,
                    use: {
                        loader: require.resolve("babel-loader"),
                        options: {
                            presets: [
                                ["@babel/preset-env", { targets: { node: nodeVersion } }]
                            ]
                        }
                    }
                }
            ]
        },
        plugins: [
            new webpack_.DefinePlugin({
                "process.env.NODE_ENV": JSON.stringify(process.env.NODE_ENV)
            })
        ]
    };
    if (!customConfig) {
        return defaultConfig;
    }
    var config = require(path.resolve(CWD$1, customConfig));
    var mergedConfig;
    if (typeof config === "function") {
        mergedConfig = config(defaultConfig);
    }
    else {
        mergedConfig = merge.smart(defaultConfig, config);
    }
    return mergedConfig;
}

function purgeRequireCache(compiler, done) {
    // missing types for webpack.Compiler
    var watchFileSystem = compiler.watchFileSystem;
    var watcher = watchFileSystem.watcher || watchFileSystem.wfs.watcher;
    // list of changed files in compiler.hooks.watchRun
    // https://stackoverflow.com/questions/43140501/can-webpack-report-which-file-triggered-a-compilation-in-watch-mode
    var changedFiles = Object.keys(watcher.mtimes);
    var _loop_1 = function (id) {
        // if node's require cache contains an entry with the filename
        // property included in `changedFiles`, remove it from cache
        // https://nodejs.org/api/modules.html#modules_require_cache
        if (changedFiles.some(function (file) { return require.cache[id].filename.includes(file); })) {
            delete require.cache[id];
        }
    };
    for (var _i = 0, _a = Object.keys(require.cache); _i < _a.length; _i++) {
        var id = _a[_i];
        _loop_1(id);
    }
    return done();
}

// bad typings
var MemoryFS = require("memory-fs");
// typescript namespace import doesn't work with Rollup
var webpack = webpack_;
function build(_a) {
    var callback = _a.callback, _b = _a.dev, dev = _b === void 0 ? false : _b, entry = _a.entry, exclude = _a.exclude, include = _a.include, nodeVersion = _a.node, targetDir = _a.target, _c = _a.watch, watch = _c === void 0 ? false : _c, customConfig = _a.webpackConfig;
    return new Promise(function (resolve, reject) {
        var entries = getEntries(entry, include, exclude);
        try {
            var config = getWebpackConfig({
                customConfig: customConfig,
                dev: dev,
                entries: entries,
                nodeVersion: nodeVersion,
                targetDir: targetDir
            });
            var compiler = webpack(config);
            if (dev) {
                var memoryFs = new MemoryFS();
                compiler.outputFileSystem = memoryFs;
                compiler.hooks.watchRun.tapAsync("lambda-dev", purgeRequireCache);
            }
            watch
                ? compiler.watch({}, getBuildCallback(entries, resolve, reject, callback))
                : compiler.run(getBuildCallback(entries, resolve, reject, callback));
        }
        catch (err) {
            console.error(chalk.grey("λ-dev") + " " + chalk.red("Build error:"), err);
            reject({ entries: entries, error: err });
        }
    });
}

// typescript namespace import doesn't work with Rollup
var requireFromString = requireFromString_;
var prefix$1 = chalk.grey("λ-dev");
var handleError = function (error, res) {
    res.status(500);
    res.send("Function invocation failed: " + error);
    return console.error(prefix$1 + " " + chalk.red("Function invocation failed:"), error);
};
var createCallback = function (res) {
    return function callback(error, lambdaRes) {
        if (error || !lambdaRes) {
            return handleError(error, res);
        }
        res.status(lambdaRes.statusCode);
        for (var key in lambdaRes.headers) {
            res.setHeader(key, lambdaRes.headers[key]);
        }
        res.send(lambdaRes.isBase64Encoded
            ? Buffer.from(lambdaRes.body, "base64")
            : lambdaRes.body);
    };
};
function promiseCallback(promise, callback) {
    return __awaiter(this, void 0, void 0, function () {
        var res, error_1;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!promise)
                        return [2 /*return*/];
                    if (typeof callback !== "function")
                        return [2 /*return*/];
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, promise];
                case 2:
                    res = _a.sent();
                    callback(null, res);
                    return [3 /*break*/, 4];
                case 3:
                    error_1 = _a.sent();
                    callback(error_1, null);
                    return [3 /*break*/, 4];
                case 4: return [2 /*return*/];
            }
        });
    });
}
var createHandler = function (source, filename) {
    return function handler(req, res) {
        var lambda = requireFromString(source, filename);
        var event = {
            path: req.path,
            httpMethod: req.method,
            queryStringParameters: req.query,
            headers: req.headers,
            body: req.body
        };
        var callback = createCallback(res);
        var promise = lambda.handler(event, {}, callback);
        promiseCallback(promise, callback).catch();
    };
};

// typescript namespace import doesn't work with Rollup
var createExpressApp = express;
var prefix$2 = chalk.grey("λ-dev");
// handlers for all routes
var routeHandlers = {};
var getPath = function (basePath) {
    var path$$1 = basePath.startsWith("/") ? basePath : "/" + basePath;
    return path$$1.endsWith("/") ? path$$1.replace(/\/$/, "") : path$$1;
};
function createServer(_a) {
    var basePath = _a.basePath, _b = _a.dev, dev = _b === void 0 ? true : _b, entry = _a.entry, exclude = _a.exclude, include = _a.include, nodeVersion = _a.node, port = _a.port, _c = _a.watch, watch = _c === void 0 ? true : _c, customConfig = _a.webpackConfig;
    return __awaiter(this, void 0, void 0, function () {
        var firstRun, app, path$$1;
        return __generator(this, function (_d) {
            switch (_d.label) {
                case 0:
                    firstRun = true;
                    app = createExpressApp();
                    app.use(bodyParser.raw());
                    app.use(bodyParser.text({ type: "*/*" }));
                    path$$1 = getPath(basePath);
                    return [4 /*yield*/, build({
                            callback: function (_a) {
                                var entries = _a.entries, modules = _a.modules;
                                var _loop_1 = function (lambda) {
                                    var requestPath = path$$1 + lambda.requestPath;
                                    var source = modules.find(function (build$$1) {
                                        return build$$1.identifier.includes(lambda.file);
                                    }).source;
                                    // set new handler for requestPath
                                    routeHandlers[requestPath] = createHandler(source, lambda.file);
                                    if (firstRun) {
                                        // create route that calls handler by requestPath
                                        app.all(requestPath, function (req, res, next) {
                                            return routeHandlers[requestPath](req, res, next);
                                        });
                                        console.log(prefix$2 + " Serving Function " + chalk.green("http://localhost:" + port + requestPath));
                                    }
                                };
                                for (var _i = 0, entries_1 = entries; _i < entries_1.length; _i++) {
                                    var lambda = entries_1[_i];
                                    _loop_1(lambda);
                                }
                                firstRun = false;
                            },
                            dev: dev,
                            entry: entry,
                            include: include,
                            exclude: exclude,
                            node: nodeVersion,
                            webpackConfig: customConfig,
                            watch: watch
                        })];
                case 1:
                    _d.sent();
                    return [2 /*return*/, app];
            }
        });
    });
}
function serve(_a) {
    var basePath = _a.basePath, entry = _a.entry, exclude = _a.exclude, include = _a.include, nodeVersion = _a.node, port = _a.port, watch = _a.watch, customConfig = _a.webpackConfig;
    return __awaiter(this, void 0, void 0, function () {
        var app;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0: return [4 /*yield*/, createServer({
                        basePath: basePath,
                        entry: entry,
                        exclude: exclude,
                        include: include,
                        node: nodeVersion,
                        port: port,
                        watch: watch,
                        webpackConfig: customConfig
                    })];
                case 1:
                    app = _b.sent();
                    return [2 /*return*/, app.listen(port, function (err) {
                            if (err) {
                                console.error(prefix$2 + " " + chalk.red("Serve error:"), err);
                                throw err;
                            }
                            else {
                                console.log(prefix$2 + " Serving...");
                            }
                        })];
            }
        });
    });
}

yargs.demandCommand(1)
    .option("include", {
    alias: "i",
    describe: "Additional glob for including Lambda Functions",
    type: "array"
})
    .option("exclude", {
    alias: "e",
    describe: "Glob for excluding Lambda Functions from the included set",
    type: "array"
})
    .option("node", {
    alias: "n",
    default: "8.10",
    description: "Target Node.js version, used with @babel/preset-env",
    type: "string"
})
    .option("webpack-config", {
    alias: "w",
    description: "Relative path to optional Webpack configuration. The default export should be an object that is merged with the default webpack config, or a function that receives the default configuration and returns a valid webpack configuration.",
    type: "string"
})
    .option("watch", {
    description: "Control the webpack watch mode. Defaults to true for serve and false for build.",
    type: "boolean"
})
    .command({
    command: "serve <entry>",
    describe: "Lambda Development Server",
    builder: function (yargs$$1) {
        return yargs$$1
            .positional("entry", {
            describe: "Entry glob for Lambda functions"
        })
            .option("basePath", {
            alias: "b",
            default: "/",
            description: "Base pathname prepended to function filenames",
            type: "string"
        })
            .option("port", {
            alias: "p",
            default: 9000,
            description: "Port for development server",
            type: "number"
        });
    },
    handler: serve
})
    .command({
    command: "build <entry> <target>",
    describe: "Bundle Lambda functions",
    builder: function (yargs$$1) {
        return yargs$$1
            .positional("entry", {
            describe: "Relative path to entry file or directory of functions",
            type: "string"
        })
            .positional("target", {
            describe: "Relative path to target directory for bundled functions",
            type: "string"
        });
    },
    handler: build
}).argv;
